#include "prjlibs-c/standards.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "skalibs/include/stddjb.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/types.h"
#include "prjlibs-c/vec.h"
#include "svclean.h"

char const* PROG="supervise-wrapper";

int main(int argc, char** argv) {
  VEC_T(char) path=VEC_0(char);
  size_t len;
  struct stat statbuf;
  type_bool const islog=(fstat(STDIN_FILENO, &statbuf)==0 &&
                         S_ISFIFO(statbuf.st_mode));

  if (argc<3 || argv[argc-1][0]=='\0') DIE_USAGE(" supervise dir");
  if (VEC_PUSHS(&path, argv[argc-1], 1)!=0 ||
      VEC_PUSHA(char, &path, SVCLEAN, 1)!=0)
    DIE0(alloc);
  len=VEC_LEN(char, &path);

  if (VEC_PUSHA(char, &path, SHUTDOWN, 0)!=0) DIE0(alloc);
  if (stat(VEC_ELTS(char, &path), &statbuf)==0) {
    if (!islog) _exit(0);
    if (ndelay_off(STDIN_FILENO)==-1) DIE1(setfl, "standard input");
    {
      iopause_fd iop;
      struct taia deadline, stamp;
      iop.fd=STDIN_FILENO;
      iop.events=IOPAUSE_READ;
      taia_now(&deadline);
      taia_now(&stamp);
      iopause(&iop, 1, &deadline, &stamp);
      if ((iop.revents&IOPAUSE_READ)==0) _exit(0);
    }
    if (ndelay_on(STDIN_FILENO)==-1) DIE1(setfl, "standard input");
  } else
    if (islog)
      if (ndelay_off(STDIN_FILENO)==-1) DIE1(setfl, "standard input");

  VEC_TRUNC(char, &path, len);
  if (VEC_PUSHA(char, &path, "/redirect_2_1", 0)!=0) DIE0(alloc);
  if (stat(VEC_ELTS(char, &path), &statbuf)==0)
    if (dup2(STDOUT_FILENO, STDERR_FILENO)!=STDERR_FILENO) DIE0(dup2);

  VEC_TRUNC(char, &path, len);
  if (VEC_PUSHA(char, &path, "/keep_", 1)!=0) DIE0(alloc);
  len=VEC_LEN(char, &path);
  if (VEC_PUSHA(char, &path, STR(FILENO_PIPERD), 0)!=0) DIE0(alloc);
  if (stat(VEC_ELTS(char, &path), &statbuf)!=0) MUST_CLOSE(FILENO_PIPERD);
  VEC_TRUNC(char, &path, len);
  if (VEC_PUSHA(char, &path, STR(FILENO_NEWERR), 0)!=0) DIE0(alloc);
  if (stat(VEC_ELTS(char, &path), &statbuf)!=0) MUST_CLOSE(FILENO_NEWERR);
  VEC_TRUNC(char, &path, len);
  if (VEC_PUSHA(char, &path, STR(FILENO_NEWOUT), 0)!=0) DIE0(alloc);
  if (stat(VEC_ELTS(char, &path), &statbuf)!=0) MUST_CLOSE(FILENO_NEWOUT);

  ++argv;
  pathexec_run(argv[0], (char const**)argv, environ);
  DIE1(exec, argv[0]);
}
