#include "prjlibs-c/standards.h"
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "skalibs/include/stddjb.h"
#include "prjlibs-c/diewarn.h"
#include "prjlibs-c/vec.h"
#include "svclean.h"

char const* PROG="svclean-shutdown";

static int status=0;
#define SETSTATUS(op) do { \
  if (status==0 || DIESTAT(#op)!=111) status=DIESTAT(#op); \
} while (0)
#define STATDIE1(op, x0) do { \
  SETSTATUS(op); \
  WARN1(op, (x0)); \
  _exit(status); \
} while (0)

static char const* servicesdir;
static size_t servicesdirlen;
static VEC_T(char) path=VEC_0(char);
static struct taia one;
static struct taia six;

static void pathadd(char const* x) {
  if (VEC_PUSHS(&path, x, 0)!=0) {
    SETSTATUS(alloc);
    WARN0(alloc);
    _exit(status);
  }
  VEC_POP1(char, &path);
}

typedef struct srvlist srvlist;
struct srvlist {
  struct taia wake_at;
  srvlist* next;
  struct taia* sleep_time;
  char const* svc_command;
  char const* log;
  unsigned char next_op;
  char name[1];
};
#define srvlist_size(len) (offsetof(srvlist, name)+(len))

typedef void srvop_t(srvlist*);
static srvop_t op_create_shutdown;
static srvop_t op_sleep;
static srvop_t op_svc;
static srvop_t* const ops[]={
  op_create_shutdown, op_sleep, op_svc, op_sleep,
  op_create_shutdown, op_sleep, op_svc, op_sleep,
  op_svc
};
static unsigned int const ops_done=(sizeof ops)/(sizeof ops[0]);

static void op_create_shutdown(srvlist* service) {
  int fd;

  VEC_TRUNC(char, &path, servicesdirlen);
  pathadd(service->name);
  pathadd(service->log);
  pathadd(SVCLEAN);
  if (mkdir(VEC_ELTS(char, &path), 0755)!=0 && errno!=EEXIST) {
    if (errno!=ENOENT) {
      SETSTATUS(mkdir);
      WARN1(mkdir, VEC_ELTS(char, &path));
    }
    service->next_op=ops_done;
    return;
  }

  pathadd(SHUTDOWN);
  fd=open_append(VEC_ELTS(char, &path));
  if (fd<0) {
    SETSTATUS(open);
    WARN1(open, VEC_ELTS(char, &path));
    service->next_op=ops_done;
    return;
  }
  if (fd_close(fd)!=0) STATDIE1(close, VEC_ELTS(char, &path));

  taia_now(&service->wake_at);
  taia_add(&service->wake_at, &service->wake_at, &one);
  service->svc_command="dx";
  service->next_op++;
}

static void op_sleep(srvlist* service) {
  struct taia now;
  taia_now(&now);
  if (taia_less(&now, &service->wake_at)) return;
  service->next_op++;
}

static void op_svc(srvlist* service) {
  int fd;

  VEC_TRUNC(char, &path, servicesdirlen);
  pathadd(service->name);
  pathadd(service->log);
  pathadd("/supervise/control");
  fd=open_write(VEC_ELTS(char, &path));
  if (fd<0) {
    if (errno==ENOENT || errno==ENXIO) {
      service->next_op++;
      if (service->next_op!=ops_done) {
        taia_now(&service->wake_at);
        taia_add(&service->wake_at, &service->wake_at, service->sleep_time);
        service->sleep_time=&six;
        service->log="/log";
      }
      return;
    }
    SETSTATUS(open);
    WARN1(open, VEC_ELTS(char, &path));
    service->next_op=ops_done;
    return;
  }

  if (service->svc_command!=NULL) {
    size_t len=strlen(service->svc_command);

    if (ndelay_off(fd)==-1) {
      SETSTATUS(setfl);
      WARN1(setfl, VEC_ELTS(char, &path));
      if (fd_close(fd)!=0) STATDIE1(close, VEC_ELTS(char, &path));
      service->next_op=ops_done;
      return;
    }

    if (allwrite(fd, service->svc_command, len)<len) {
      SETSTATUS(write);
      WARN1(write, VEC_ELTS(char, &path));
      if (fd_close(fd)!=0) STATDIE1(close, VEC_ELTS(char, &path));
      service->next_op=ops_done;
      return;
    }
  }

  if (fd_close(fd)!=0) STATDIE1(close, VEC_ELTS(char, &path));

  service->svc_command=(*service->log=='\0'? NULL: "x");
}

int main(int argc, char** argv) {
  DIR* dirh;
  srvlist* services, * service, ** next;
  struct taia zero;

  if (argc!=2 || argv[1][0]=='\0') DIE_USAGE(" /service\n");

  servicesdir=argv[1];
  pathadd(servicesdir);
  if (*VEC_ELT(char, &path, VEC_LEN(char, &path)-1)!='/') pathadd("/");
  servicesdirlen=VEC_LEN(char, &path);
  dirh=opendir(servicesdir);
  if (dirh==NULL) STATDIE1(open, servicesdir);

  next=&services;
  for (;;) {
    struct dirent const* de;
    size_t namelen, size;

    errno=0;
    de=readdir(dirh);
    if (de==NULL) {
      if (errno==0) break;
      STATDIE1(read, servicesdir);
    }

    if (de->d_name[0]=='.') continue;

    namelen=strlen(de->d_name)+1u;
    size=srvlist_size(namelen);
    if (size<namelen) DIE_OVERFLOW();

    service=malloc(size);
    if (service==NULL) {
      SETSTATUS(alloc);
      WARN0(alloc);
      _exit(status);
    }

    memcpy(service->name, de->d_name, namelen);
    service->log="";
    service->next_op=0;
    service->sleep_time=&one;

    *next=service;
    next=&service->next;
  }
  closedir(dirh);
  *next=NULL;

  if (services==NULL) _exit(status);

  taia_uint(&zero, 0u);
  taia_uint(&one, 1u);
  taia_uint(&six, 6u);

  for (;;) {
    /* are all services completely shut down? */
    unsigned int all_done=1;
    /* are any either polling or ready to do something? */
    unsigned int any_not_sleeping=0;
    /* are none waiting for a certain interval after some event? */
    unsigned int none_sleeping=1;
    /* are all either waiting or polling? */
    unsigned int do_sleep=1;
    /* when will we next have something to do? */
    struct taia wake_at;

    for (service=services; service!=NULL; service=service->next) {
      unsigned char const last_op=service->next_op;
      if (last_op==ops_done) continue;
      ops[service->next_op](service);
      if (service->next_op==ops_done) continue;
      all_done=0;
      if (ops[service->next_op]==op_sleep) {
        if (none_sleeping || taia_less(&service->wake_at, &wake_at))
          taia_add(&wake_at, &service->wake_at, &zero);
        none_sleeping=0;
      } else {
        any_not_sleeping=1;
        if (service->next_op!=last_op) do_sleep=0;
      }
    }

    if (all_done) _exit(status);

    if (do_sleep) {
      iopause_fd iop;
      struct taia now, now_plus_one;
      taia_now(&now);
      taia_add(&now_plus_one, &now, &one);
      if (none_sleeping ||
          (any_not_sleeping && taia_less(&now_plus_one, &wake_at)))
        /* at least one is polling; we won't sleep too long */
        iopause(&iop, 0, &now_plus_one, &now);
      else
        iopause(&iop, 0, &wake_at, &now);
    }
  }
}
